#ifndef BWIDGETS_EXAMPLES_RUN_HPP
#define BWIDGETS_EXAMPLES_RUN_HPP

#include <functional>
#include <iostream>
#include <stack>

#include <SDL2/SDL.h>
#include <SDL2/SDL_ttf.h>

#include <basic_widgets/core/font.hpp>
#include <basic_widgets/core/renderer.hpp>
#include <basic_widgets/w/aligned_layout.hpp>
#include <basic_widgets/w/base/widget.hpp>

template<typename T>
concept WidgetType = std::derived_from<T, bwidgets::Widget>;

template<WidgetType W>
void run_example(std::function<void(W*, bwidgets::Font*, int, int)> setup, int w = 3,
                 int h = 3)
{
    std::stack<std::function<void()>> cleaners;
    try {
        bwidgets::SDLError::success_or_throw(SDL_Init(SDL_INIT_VIDEO), __FILE__,
                                             __FUNCTION__, __LINE__);
        cleaners.emplace([]() { SDL_Quit(); });
        bwidgets::SDLError::success_or_throw(TTF_Init(), __FILE__, __FUNCTION__,
                                             __LINE__);
        cleaners.emplace([]() { TTF_Quit(); });

        const bwidgets::Size size_init {854, 480};
        auto*                font {new bwidgets::Font(bwidgets::Font::find("Monospace"),
                                                      16)}; // NOLINT(readability-magic-numbers)
        cleaners.emplace([font]() { delete font; });
        auto* win {SDL_CreateWindow("basic_widgets example", SDL_WINDOWPOS_CENTERED,
                                    SDL_WINDOWPOS_CENTERED, size_init.w, size_init.h,
                                    SDL_WINDOW_SHOWN | SDL_WINDOW_RESIZABLE
                                      | SDL_WINDOW_UTILITY)};
        cleaners.emplace([win]() { SDL_DestroyWindow(win); });
        auto* renderer {new bwidgets::Renderer(win, -1, SDL_RENDERER_ACCELERATED)};
        cleaners.emplace([renderer]() { delete renderer; });
        renderer->blend_mode(SDL_BLENDMODE_BLEND);

        auto* layout {
          new bwidgets::AlignedLayout(bwidgets::AlignedLayout::Alignment::HORIZONTAL)};
        cleaners.emplace([layout]() { delete layout; });

        for (auto x = 0; x < w; x++) {
            auto* col =
              new bwidgets::AlignedLayout(bwidgets::AlignedLayout::Alignment::VERTICAL);
            for (auto y = 0; y < h; y++) {
                auto* widget = new W();
                setup(widget, font, x, y);
                col->add_widget(widget);
            }
            layout->add_widget(col);
        }

        layout->renderer(renderer);
        layout->viewport({0, 0, size_init.w, size_init.h});

        bool quit {false};
        while (!quit) {
            SDL_Event ev;
            while (SDL_PollEvent(&ev) != 0) {
                switch (ev.type) {
                    case SDL_QUIT:
                        quit = true;
                        break;
                    case SDL_WINDOWEVENT:
                        if (ev.window.event == SDL_WINDOWEVENT_SIZE_CHANGED) {
                            auto size = renderer->output_size();
                            layout->viewport({0, 0, size.w, size.h});
                        }
                        break;
                }
                layout->handle_event(ev);
            }

            // NOLINTNEXTLINE(readability-magic-numbers)
            renderer->draw_color({50, 60, 70, SDL_ALPHA_OPAQUE});
            renderer->clear();

            layout->render();
            renderer->present();
        }
        while (!cleaners.empty()) cleaners.top()(), cleaners.pop();
    } catch (const bwidgets::SDLError& e) {
        std::cerr << "Error: " << e.file << ":" << e.func << ":" << e.line << ": "
                  << e.what << std::endl;
        while (!cleaners.empty()) cleaners.top()(), cleaners.pop();
    }
}

#endif
