#include <functional>
#include <stack>

#include <fontconfig/fontconfig.h>

#include <basic_widgets/core/font.hpp>
#include <basic_widgets/core/type/fc_error.hpp>
#include <basic_widgets/core/type/sdl_error.hpp>

namespace bwidgets
{
    Font::Font(TTF_Font* f)
      : OpaqueStruct(f),
        ascent(TTF_FontAscent(f)),
        descent(TTF_FontDescent(f)),
        faces(TTF_FontFaces(f)),
        family_name(TTF_FontFaceFamilyName(f)),
        fixed_width(TTF_FontFaceIsFixedWidth(f) != 0),
        height(TTF_FontHeight(f)),
        line_skip(TTF_FontLineSkip(f)),
        style_name(TTF_FontFaceStyleName(f))
    {}

    Font::Font(const std::string& file, int size)
      : Font(ptr_or_throw<FCError>(TTF_OpenFont(file.c_str(), size), __FILE__,
                                   __FUNCTION__, __LINE__, "failed to open font"))
    {}

    Font::~Font()
    {
        TTF_CloseFont(_c_pod);
    }

    auto Font::hinting() -> Font::Hinting
    {
        return (Hinting)TTF_GetFontHinting(_c_pod);
    }

    auto Font::hinting(Font::Hinting h) -> Font*
    {
        TTF_SetFontHinting(_c_pod, (int)h);
        return this;
    }

    auto Font::kerning() -> bool
    {
        return TTF_GetFontKerning(_c_pod) != 0;
    }

    auto Font::kerning(bool allowed) -> Font*
    {
        TTF_SetFontKerning(_c_pod, static_cast<int>(allowed));
        return this;
    }

    auto Font::outline() -> int
    {
        return TTF_GetFontOutline(_c_pod);
    }

    auto Font::outline(int size) -> Font*
    {
        TTF_SetFontOutline(_c_pod, size);
        return this;
    }

    auto Font::render(RenderMode m, const std::string& str, const Color& fg,
                      const Color& bg) -> SDL_Surface*
    {
        std::function<SDL_Surface*()> renderer;
        const char*                   c_str = str.empty() ? " " : str.c_str();

        switch (m) {
            case RenderMode::BLENDED:
                renderer = [&fg, c_str, this]() {
                    return TTF_RenderUTF8_Blended(_c_pod, c_str, fg());
                };
                break;
            case RenderMode::SHADED:
                renderer = [&bg, &fg, c_str, this]() {
                    return TTF_RenderUTF8_Shaded(_c_pod, c_str, fg(), bg());
                };
                break;
            case RenderMode::SOLID:
                renderer = [&fg, c_str, this]() {
                    return TTF_RenderUTF8_Solid(_c_pod, c_str, fg());
                };
                break;
        }

        return ptr_or_throw<SDLError>(renderer(), __FILE__, __FUNCTION__, __LINE__);
    }

    auto Font::style() -> uint8_t
    {
        return TTF_GetFontStyle(_c_pod);
    }

    auto Font::style(uint8_t s) -> Font*
    {
        TTF_SetFontStyle(_c_pod, s);
        return this;
    }

    auto Font::text_size(const std::string& str) -> Size
    {
        Size s {};
        TTF_SizeUTF8(_c_pod, str.c_str(), &s.w, &s.h);
        return s;
    }

    auto Font::find(const std::string& pat) -> std::string
    {
        std::stack<std::function<void()>> cleaners;

        FcConfig* conf = ptr_or_throw<FCError>(FcInitLoadConfigAndFonts(), __FILE__,
                                               __FUNCTION__, __LINE__, "init failed");
        cleaners.emplace([conf]() { FcConfigDestroy(conf); });

        FcPattern* pattern = nullptr;
        try {
            pattern =
              ptr_or_throw<FCError>(FcNameParse((FcChar8*)pat.c_str()), __FILE__,
                                    __FUNCTION__, __LINE__, "pattern parsing failed");
            cleaners.emplace([pattern]() { FcPatternDestroy(pattern); });

            if (FcConfigSubstitute(conf, pattern, FcMatchPattern) == FcFalse)
                throw FCError {__FILE__, __FUNCTION__, __LINE__,
                               "FcConfigSubstitute failed"};
        } catch (const std::exception& e) {
            while (!cleaners.empty()) {
                cleaners.top()();
                cleaners.pop();
            }
            throw e;
        }

        FcDefaultSubstitute(pattern);

        std::string file_path;
        FcResult    res {};
        FcPattern*  font = FcFontMatch(conf, pattern, &res);
        if (font != nullptr) {
            FcChar8* file = nullptr;
            if (FcPatternGetString(font, FC_FILE, 0, &file) == FcResultMatch) {
                file_path = (char*)file;
            }
            FcPatternDestroy(font);
        }

        while (!cleaners.empty()) {
            cleaners.top()();
            cleaners.pop();
        }

        if (file_path.empty())
            throw FCError {__FILE__, __FUNCTION__, __LINE__, "no font found"};

        return file_path;
    }
}
