#include <basic_widgets/core/renderer.hpp>
#include <basic_widgets/core/texture.hpp>

using namespace bwidgets;

Renderer::Renderer(SDL_Renderer* r) : OpaqueStruct(r), info(_info(r)) {}

Renderer::Renderer(SDL_Window* w, int index, uint32_t flags = 0)
  : Renderer(ptr_or_throw<SDLError>(SDL_CreateRenderer(w, index, flags), __FILE__,
                                    __FUNCTION__, __LINE__))
{}

Renderer::~Renderer()
{
    SDL_DestroyRenderer(_c_pod);
}

auto Renderer::blend_mode() -> SDL_BlendMode
{
    SDL_BlendMode mode {};
    SDLError::success_or_throw(SDL_GetRenderDrawBlendMode(_c_pod, &mode), __FILE__,
                               __FUNCTION__, __LINE__);

    return mode;
}

auto Renderer::blend_mode(SDL_BlendMode mode) -> Renderer*
{
    SDLError::success_or_throw(SDL_SetRenderDrawBlendMode(_c_pod, mode), __FILE__,
                               __FUNCTION__, __LINE__);

    return this;
}

auto Renderer::clear() -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderClear(_c_pod), __FILE__, __FUNCTION__,
                               __LINE__);

    return this;
}

auto Renderer::copy(Texture* t, const SDL_Rect* src, const SDL_Rect* dst) -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderCopy(_c_pod, (*t)(), src, dst), __FILE__,
                               __FUNCTION__, __LINE__);

    return this;
}

auto Renderer::draw_color() -> Color
{
    Color c;
    SDLError::success_or_throw(
      SDL_GetRenderDrawColor(_c_pod, &c().r, &c().g, &c().b, &c().a), __FILE__,
      __FUNCTION__, __LINE__);

    return c;
}

auto Renderer::draw_color(const Color& c) -> Renderer*
{
    SDLError::success_or_throw(
      SDL_SetRenderDrawColor(_c_pod, c().r, c().g, c().b, c().a), __FILE__, __FUNCTION__,
      __LINE__);

    return this;
}

auto Renderer::draw_line(const SDL_Point& a, const SDL_Point& b) -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderDrawLine(_c_pod, a.x, a.y, b.x, b.y), __FILE__,
                               __FUNCTION__, __LINE__);

    return this;
}

auto Renderer::draw_lines(const std::vector<SDL_Point>& pts) -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderDrawLines(_c_pod, pts.data(), (int)pts.size()),
                               __FILE__, __FUNCTION__, __LINE__);

    return this;
}

auto Renderer::draw_point(const SDL_Point& p) -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderDrawPoint(_c_pod, p.x, p.y), __FILE__,
                               __FUNCTION__, __LINE__);

    return this;
}

auto Renderer::draw_points(const std::vector<SDL_Point>& pts) -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderDrawPoints(_c_pod, pts.data(), (int)pts.size()),
                               __FILE__, __FUNCTION__, __LINE__);

    return this;
}

auto Renderer::draw_rect(const SDL_Rect* r) -> Renderer*
{
    auto vp = viewport();
    // Has glitch at top-left and bottom-right corner.
    // The first corner has an extra pixel at y-1 and surimpression.
    // The second corner is missing a pixel.
    if (r != nullptr)
        viewport({vp.x + r->x, vp.y + r->y, r->w - 1, r->h - 1}); // crop extra pixel
    SDLError::success_or_throw(SDL_RenderDrawRect(_c_pod, nullptr), __FILE__,
                               __FUNCTION__, __LINE__);
    // add missing pixel. works sometimes…
    if (r != nullptr) draw_point({r->w - 1, r->h - 1});
    else draw_point({vp.w - 1, vp.h - 1});

    viewport(vp);

    return this;
}

auto Renderer::draw_rects(const std::vector<SDL_Rect>& rs) -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderDrawRects(_c_pod, rs.data(), (int)rs.size()),
                               __FILE__, __FUNCTION__, __LINE__);

    return this;
}

auto Renderer::fill_rect(const SDL_Rect* r) -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderFillRect(_c_pod, r), __FILE__, __FUNCTION__,
                               __LINE__);

    return this;
}

auto Renderer::fill_rects(const std::vector<SDL_Rect>& rs) -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderFillRects(_c_pod, rs.data(), (int)rs.size()),
                               __FILE__, __FUNCTION__, __LINE__);

    return this;
}

auto Renderer::output_size() -> Size
{
    Size s {};
    SDLError::success_or_throw(SDL_GetRendererOutputSize(_c_pod, &s.w, &s.h), __FILE__,
                               __FUNCTION__, __LINE__);

    return s;
}

void Renderer::present()
{
    SDL_RenderPresent(_c_pod);
}

auto Renderer::viewport() -> SDL_Rect
{
    SDL_Rect vp;
    SDL_RenderGetViewport(_c_pod, &vp);

    return vp;
}

auto Renderer::viewport(const SDL_Rect* vp) -> Renderer*
{
    SDLError::success_or_throw(SDL_RenderSetViewport(_c_pod, vp), __FILE__, __FUNCTION__,
                               __LINE__);

    return this;
}
